@@ -6,12 +6,15 @@
 
 use crate::{MessageId, Versioned};
 use serde::de::DeserializeOwned;
+use serde::Serialize;
 use std::any::type_name;
 
 /// A data structure that contains a message-id and version fields.
 
 pub trait GroupHeader {
+    /// Retrieve the message id.
     fn msg_id(&self) -> u16;
+    /// Retrieve the message version.
     fn msg_ver(&self) -> u16;
 }
 
@@ -25,7 +28,7 @@ pub trait GroupHeader {
 // b) User needs to specify a range or list of versions
 pub trait UpgradeLatest: DeserializeOwned + Versioned {
     /// Deserialize version `ver` of the target struct, then upgrade it to the latest version.
-    fn upgrade_latest<Src>(src: &mut Src, ver: u16) -> Result<Self, Src::Error>
+    fn upgrade_latest<Src>(src: &mut Src, header: Src::Header) -> Result<Self, Src::Error>
     where
         Src: DataSource;
 }
@@ -59,7 +62,7 @@ pub trait DataSource {
     ///
     /// This is a user-defined function that will deserialize a message
     /// of type `T`.
-    fn read_message<T>(&mut self) -> Result<T, Self::Error>
+    fn read_message<T>(&mut self, header: &Self::Header) -> Result<T, Self::Error>
     where
         T: DeserializeOwned;
 
@@ -101,6 +104,41 @@ pub trait DataSource {
     }
 }
 
+/// Useful functions for `DataSource`.
+///
+/// There is a blanket implementation of this trait, so that any
+/// [`DataSource`] type can use these functions.
+pub trait DataSourceExt: DataSource {
+    /// Read a specific message type from the `DataSource`.
+    ///
+    /// This will read the message header, and if the message id matches
+    /// the type `T` that was requested, read the message.
+    /// The message will be upgraded to the latest version, and then
+    /// returned.
+    fn expect_message<T>(&mut self) -> Result<T, Self::Error>
+    where
+        T: MessageId + UpgradeLatest;
+}
+
+impl<Src> DataSourceExt for Src
+where
+    Src: DataSource,
+{
+    fn expect_message<T>(&mut self) -> Result<T, Src::Error>
+    where
+        Src: DataSource,
+        T: MessageId + UpgradeLatest,
+    {
+        let header: Src::Header = self.read_header()?;
+        if header.msg_id() == T::MSG_ID {
+            T::upgrade_latest(self, header)
+        } else {
+            // Call the user-supplied error fn
+            Err(self.unexpected_message::<T>(header.msg_id()))
+        }
+    }
+}
+
 /// A derived trait that can deserialize any message from a group.
 pub trait GroupDeserialize: Sized {
     /// Read the next message from the `DataSource`.
@@ -112,24 +150,23 @@ pub trait GroupDeserialize: Sized {
     fn read_message<Src>(src: &mut Src) -> Result<Self, Src::Error>
     where
         Src: DataSource;
+}
 
-    /// Read a specific message type from the `DataSource`.
+/// `DataSink` allows user-defined IO, deserialization, and
+/// error handling.
+///
+pub trait DataSink {
+    /// A user-defined error type.
     ///
-    /// This will read the message header, and if the message id matches
-    /// the type `T` that was requested, read the message.
-    /// The message will be upgraded to the latest version, and then
-    /// returned.
-    fn expect_message<Src, T>(src: &mut Src) -> Result<T, Src::Error>
+    /// This error type will be returned from all trait member functions.
+    /// It's probably a good idea for it to be able to represent IO errors,
+    /// deserialization errors, and "unknown message" errors.
+    type Error;
+
+    /// Write a header and message to the data sink.
+    ///
+    fn write_message<T>(&mut self, msg: &T) -> Result<(), Self::Error>
     where
-        Src: DataSource,
-        T: MessageId + UpgradeLatest,
-    {
-        let header: Src::Header = src.read_header()?;
-        if header.msg_id() == T::MSG_ID {
-            T::upgrade_latest(src, header.msg_ver())
-        } else {
-            // Call the user-supplied error fn
-            Err(src.unexpected_message::<T>(header.msg_id()))
-        }
-    }
+        T: Serialize + Versioned,
+        T::Base: MessageId;
 }
